function [xbest, fbest, xnew_history, fbest_history, efficiency] = ...
    eHSGA(f, xL, xU, eHSGAparams)
% An Elitist Harmony Search and Genetic Algorithms Hybridization
% This code has been written by H. Binol on March, 2018.
% If you have any questions please contact me: hbinol@fiu.edu

% Unpack the parameters
HMS       = eHSGAparams.HMS;
MaxImp    = eHSGAparams.MaxImp;
HMCRmax   = eHSGAparams.HMCRmax;
HMCRmin   = eHSGAparams.HMCRmin;
PARmin    = eHSGAparams.PARmin;
PARmax    = eHSGAparams.PARmax;
SCthresh  = eHSGAparams.SC;

fw        = 1.0;

% Dimension arrays
N             = length(xL);      % Number of decision variables
HM{HMS, 1}    = [];
HMnew{HMS, 1} = []; % generated harmony
F             = zeros(HMS, 1);
F2            = zeros(HMS, 1); % objective function values of generated harmony
fbest_history = zeros(MaxImp, 1);
xnew_history  = zeros(MaxImp, N);

% Evaluate the corresponding objective function values.
for j = 1:HMS
    harmony.value   = eHSGAparams.initPop(j,:);
    harmony.fitness = f(harmony.value);
    harmony.exist   = 0;
    
    HM{j}           = harmony;
    F(j)            = harmony.fitness;
end

% Evaluate the fitness for each individual in HM
[F, ind] = sort(F);
HM       = HM(ind);

% Loop through MaxImp improvisations
for j = 1:MaxImp
    
    xbest = HM{1};
    xbest.fitness;
    
    HMCR = HMCRmax - ((HMCRmax - HMCRmin) * (1 * j / MaxImp));
    
    % PAR =(PARmax - PARmin) / (MaxIter * counter) + PARmin; % alternative
    PAR = (PARmax - PARmin) * (1 * j / MaxImp) + PARmin;
    
    % Improvise a new harmony memory
    for i = 1:HMS
        % Randomly perform one of the eHSGA operation
        if rand < HMCR
            % Memory considering: randomly select a note stored in HM
            
            if rand < PAR
                % Pitch adjusting: randomly adjust the pitch slightly
                % within +/- b(i), and ensure bounds are satisfied
                
                % pitch adjust just on a one point of xbest
                cp = randperm(N, 1); % pick a random point
                
                xnew.value     = xbest.value;
                xnew.value(cp) = xnew.value(cp) + (2*rand-1) * fw;

            else
                % crossover
                cp = randperm(N-1, 1); % random crossover point
                
                % crossover with a random member
                tempV             = randperm(HMS);
                tempV(tempV == i) = [];
                
                if rand < 0.5
                    xnew.value = [HM{i}.value(1:cp) ...
                        HM{tempV(1)}.value((cp + 1):end)];
                else
                    xnew.value = [HM{tempV(1)}.value(1:cp) ...
                        HM{i}.value((cp + 1):end)];
                end
                 
            end
        else
            % Random playing: randomly select any pitch within bounds
            xnew.value = xL + (xU - xL) .* rand(1,N);
         
        end
        
        % record the new harmony
        xnew.exist                     = 0;
        xnew.fitness                   = f(xnew.value);
        F2(i)                          = xnew.fitness;
        HMnew{i}                       = xnew;
        xnew_history((j-1)*HMS + i, :) = xnew.value;
        
    end % Finished constructing a new harmony memory (HM)
    
    % HM update: merge HMnew and HM and choose the bests
    Ftemp         = [F; F2];
    HMtemp        = [HM; HMnew];
    [Ftemp, ind2] = sort(Ftemp);
    HMtemp        = HMtemp(ind2);
    
    idx           = find(ind2 == 1);
    convEff       = (idx - 1) / HMS;
    
    HM            = HMtemp(1:HMS);
    F             = Ftemp(1:HMS);
    
    % mutation operator on long-term survivals
    for k = 1:HMS
        HM{k}.exist = HM{k}.exist + 1;
        if HM{k}.exist >= SCthresh
            r          = randperm(N, 1); % mutation point
            xmut.value = HM{k}.value;
            
            xmut.value(r) = xmut.value(r) + (2*rand-1) * fw;
            xmut.fitness  = f(xmut.value);
            
            if xmut.fitness < HM{k}.fitness
                HM{k}       = xmut;
                HM{k}.exist = 0;
                
                if xmut.fitness < F(1)
                    
                    convEff    = convEff + 1/HMS;
                    
                    for t = 1:HMS
                        F(t) = HM{t}.fitness;
                    end
                    
                    [F, indm] = sort(F);
                    HM        = HM(indm);
                end
            end
        end
        
    end
    
    fw = fw * 0.99;
    
    efficiency(j)   = convEff;
    
    % Record the best function value
    fbest_history(j) = F(1);
    
end % Maximum number of improvisations reached

% Return the best harmony found
[fbest, idxbest] = min(F);
xbest            = HM{idxbest}.value;